/*
    utility for the message handling library
    Copyright (C) 1995  Steffen Kaiser

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   
*/
/* $RCSfile: msgcomp.c $
   $Locker: ska $	$Name:  $	$State: Exp $

	The message compiler.

   ----
	command line usage:

	msgComp [/!] glbMsg msgFile [comp.rsp [lib.rsp [decFile]]

	glbMsg  : global message file
	msgFile : message file to be compiled
	compLang: file, which maps the English MSGID to a language specific MSGID
	comp.rsp: compiler response file
	lib.rsp : librarian response file
	decFile : message declaration file

	The message compiler creates numerous message definition files
	"msg_####.c" containing the message string definitions and the
	header file (msgFile with the extension ".msg") containing the
	message string declarations, the error level definitions, and
	the mapping to the native language (see below).
	The /! switchs the name to "msg!####", because VAL complains about
	multiple modules. So the global messages can be generated with
	this switch.

   ----
   	file formats:

   	msgFile: with extension ".msg" ==> the real messages

   	Each message consists of a message head and a mssage body. The message
   	head is the whole line of the format: ">>>>>>?_name [#]\n". The message
   	body consists of all lines up to the next message header or the EOF.

   	name :: is the name of the message in English
   	?    :: is the message group and must be a single upper-cased letter
   	#    :: error level (for ? == 'E' only)

	The message body may contain special strings: "%-%*%", where '*' stands
	for a comma delimited list of the following ANSI keywords:
		cls			clear screen
		clreol		clear from cursor to end of line
		
		home		cursor to upper left corner
		up #		cursor up # rows		\
		down #		cursor down # rows		 | neither scroll nor change
		left #		cursor left # columns	 | of the other coordiante
		right #		cursor right # columns	/
		goto # #	move cursor to position row,column (one based!!)

		save		save one cursor position
		restore		restore saved cursor position

	  colour related keywords:
		bright		set attribute: foreground colour highlighted
		blink		set attribute: blinking character
		inverse		set attribute: inverse
		underline	set attribute: underline
		hidden		set attribute: hidden
		normal		unset all attributes
		black		foreground colour: black 
		red			foreground colour: red 
		green		foreground colour: green 
		yellow		foreground colour: yellow 
		blue		foreground colour: blue 
		magenta		foreground colour: magenta 
		cyan		foreground colour: cyan 
		white		foreground colour: white 

		Those keywords may be combined in any order WITHOUT a delimiting
		comma. If more than one colour is given, the first is treated as
		a foreground colour, all others as background colours.

	These keywords must be defined in the global message file within
	message group 'I' == internal.

	To define custom strings use "%#-%*", with # the number of bytes and
	* as # bytes. 

	Because the message body is transformed into a C string definition,
	any non-printable characters are replaced by their C-style and already
	enterred C-style is interpreted by the C compiler, the last '\n' is
	stripped, e.g.:

>>>>>>M_dummy
	line\\	1
	line\r	2
>>>>>>

	is transformed into:

	"\tline\\\t1\n\tline\r\t2"


	glbMsg: global message file (same syntax as msgFile)

	must define all internal strings



	msgFile: with extension ".map" ==> MSGID mapping for programming

	each line defines the map of the English message name into the
	destination language. Not all names need to be mapped, e.g.:
	I_ja I_yes
	I_nein I_no
	I_weiss I_white




	comp.rsp: compiler response file

	each line contains one message definition file name, e.g.:

	msg00001.c
	msg00002.c




	lib.rsp: librarian response file

	each line contains a librarian statement, e.g.:

	+- msg00001.obj &
	+- msg00002.obj


	decFile: message declaration file

	This file is updated to match the current defined messages. This
	file MUST already exist, and is backuped to "*.h__", because it
	should be a C-header file.
	From the backup all lines are copied into the new decFile, until
	either end of file or a line starting with "DCL_MSG(" (no leading
	whitespaces!) is encountered. 
	Then for each message header of msgFile the corresponding "DCL_MSG()"
	line is written into the decFile.
	When the msgFile is read, the rest of the backup is appended cutting
	out all lines starting with "DCL_MSG(". Note: #ifdef's etc. are not
	processed, therefore conditional message declaration files must be
	updated manually!

	
	Known bugs:
		Most compilers have problems with '\0' characters within
		the IN stream. Therefore, it's not generally possible to
		place them e.g. in the %#-%??? special. Also, because of
		the text mode reading '\x1a' (^Z) is not allowed as a
		legal character.

		The string attached to the %#-%??? special is not
		processed in any way, especially no C-style syntax is
		translated here.

   $Log: msgcomp.c $
   Revision 1.7  1995/05/26 22:12:49  ska
   chg: msg def file names into "msg_####", where #### is to the base 36
   add: /! switch to use "msg!####" instead

   Revision 1.6  1995/05/26 20:58:20  ska
   add: Update of the msg declaration file.

   Revision 1.5  1995/04/21 04:35:09  ska
   version 1.0

   Revision 1.4  1995/04/21 03:51:23  ska
   chg: one byte alignment.

   Revision 1.3  1995/04/19 09:06:46  ska
   bugfix: Efopen(?,?,NULL)
   bugfix: translateANSI() non-colour keyword

   Revision 1.2  1995/04/11 21:50:03  ska
   BC hack.

   Revision 1.1  1995/04/10 23:20:00  ska
   Initial revision
   	Supported compilers:
   		+ Micro-C updated v3.13

*/

#include <stdio.h>
#ifndef _MICROC_
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#endif
#include <portable.h>
#include <freedos.h>

#ifdef __TURBOC__
#pragma hdrstop
#endif

//#define ERROR_MAKING_STATE
#include "yerror.h"

#define NUL '\0'
#define TRIGGER "DCL_MSG("	/* begin of a line marking a message declaration */
#define BUF 2048 	/* buffer size while copying a file */

#define IN_PREFIX "I_"	/* prefix of internal used messages */

#define IN_white 0
#define IN_cls 1
#define IN_clreol 2
#define IN_home 3
#define IN_up 4
#define IN_down 5
#define IN_left 6
#define IN_right 7
#define IN_goto 8
#define IN_save 9
#define IN_restore 10
#define IN_bright 11
#define IN_blink 12
#define IN_inverse 13
#define IN_underline 14
#define IN_hidden 15
#define IN_normal 16
#define IN_black 17
#define IN_red 18
#define IN_green 19
#define IN_yellow 20
#define IN_blue 21
#define IN_magenta 22
#define IN_cyan 23

#define IN_Nr 24	/* # of internal names */

#ifndef lint
static char const rcsid[] = 
	"$Id: msgcomp.c 1.7 1995/05/26 22:12:49 ska Exp ska $";
#endif


/* opening of all message definition files containing errors */
const char msgErrOpening[] = "\
#ifndef _MICROC_\n\
#include <algnbyte.h>\n\
#endif\n\
#include <portable.h>\n\
static unsigned char errNr = %u;\n\
const unsigned char %s[] = \"\\\n";

/* opening of all message definition files don't contain errors */
const char msgOpening[] = "\
#include <portable.h>\n\
const unsigned char %s[] = \"\\\n";

/* closing of all message definition files */
const char msgClosing[] = "\
\";\n";

char *internal[] = {
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

	/* fixed extensions for: */
const char extMsg[] = ".msg";	/* message file (global & local) */
const char extBak[] = ".h__";	/* message decl file backup */

char msgDefFile[] = "%smsg_%04s%s";	/* name of the message definition files */
#define padDefFile 5		/* offset of '_' */

char *msgFile; /* extension-less local message file; with space for ext */
char *glbMsgFile; /* extension-less global message file; with space for ext */
char *msgBody = NULL;	/* current message body */

/* unsigned msgDefFileNr = 0;*/	/* number of message definition files */
char zeroDefFileNr[] = "0000";
char msgDefFileNr[] = "0000";
char msgDefFileStr[13];

unsigned errNr = 0;	/* error level of current message header */
unsigned lineNr = 0;	/* keep it NOT (long) because of Micro-C */

byte *getmem(int len)
{	byte *h;

	if((h = malloc(len)) == NULL)
		error(E_noMem);

	return h;
}



#define addStr(a,b) _addStr(&(a),(b))
char *_addStr(char **str, char *s)
/* conact string s onto dynamic string str */
{
	if(s == NULL) return *str;
	if(*str == NULL) {
		if(*str = strdup(s))
			return *str;
		fatal(E_noMem);
	}
	if((*str = realloc(*str, strlen(*str) + strlen(s) + 1)) == NULL)
		fatal(E_noMem);
#ifdef _MICROC_
	stpcat(*str, s);
	return *str;
#else
	return strcat(*str, s);
#endif
}

FILE *Efopen(char *fname, const char *mode, const char *ext)
/* Open file fname with the extension ext and mode mode.

   On error emit error message.
*/
{	FILE *fp;
	char *p;
	/* append extension */
	if(ext) 
		stpcpy(p = strchr(fname, '\0'), ext);

	/* open file & emit error on failure */
	if((fp = fopen(fname, mode)) == NULL)
		fatal(E_openFile, fname);

	if(ext)
		*p = '\0';	/* strip extension */
	return fp;
}

void Efclose(FILE *fp, char *fnam, const char *ext)
{	/* ensure, fp->buffer is empty, so ferror() is up to date */
	fflush(fp);

	if(ferror(fp)) {
		/* append extension */
		if(ext) 
			stpcat(fnam, ext);
		fatal(E_writeFile, fnam);
	}

	fclose(fp);
}



char *mkFname(char *fname)
/* duplicate fname
   strip extension from fname
   leave enough space behind duplicated fname to add an extension */
{	char *h, *p;

	p = stpcpy(h = getmem(strlen(fname) + 5),	/* 4 chars extension, 1 NUL */
		fname);
	while(p > h && *p != '\\' && *p != '/' && *p != ':')
		if(*--p == '.') {
			*p = NUL;
			break;
		}

	return h;
}

char *readMsgBody(FILE *fp)
/*	read next lines, until EOF or an message header encountered */
{	char line[80];
	char *p;
	int cnt;
	int c;

	free(msgBody);
	msgBody = NULL;
	do {
		p = line - 1;
		cnt = 1;
		++lineNr;
		while((c = fgetc(fp)) != EOF && (*++p = c) != '\n')
			if(p > line + 78) { 
				p[1] = NUL;
				addStr(msgBody, line);
				p = line - 1;
			} 
			else if(c == '>') {
				if(cnt && ++cnt == 7) {	/* message header found */
					p[-5] = NUL;	/* cut out '>>>>>>' */
					c = EOF;	/* stop reading */
					break;
				}
			}
			else cnt = 0;	/* no message header */

		p[1] = NUL;
		addStr(msgBody, line);
	} while(c != EOF);

	/* strip last '\n' */
	if(*(p = strchr(msgBody, NUL) - 1) == '\n')
		*p = NUL;
	return msgBody;
}

char *readMsgHeader(FILE *fp)
/* treat the next line as a message header, the '>>>>>>" are already
   read */
{	static char msgHeader[40];
	char *p;
	int c;

	*(p = msgHeader) = NUL;
	while((isalnum(c = fgetc(fp)) || c == '_') && p < &msgHeader[39])
		*p++ = c;
	*p = NUL;

	while(c != '\n' && isspace(c))
		c = fgetc(fp);

	if(*msgHeader == 'E') { /* scan error level */
		if(!isdigit(c))
			error(E_msgFileHeadErrLvl, lineNr);
		errNr = 0;
		do errNr = errNr * 10 + c - '0';
		while(isdigit(c = fgetc(fp)));

		while(c != '\n' && isspace(c))
			c = fgetc(fp);
	}

	if(*msgHeader == NUL && c == EOF)
		return NULL;

	if(*msgHeader < 'A' || *msgHeader > 'Z' || msgHeader[1] != '_' 
	  || c != '\n' && c != EOF)
		error(E_msgFileHeadStruc, lineNr);

	readMsgBody(fp);	/* scan in the body and supply it in the global msgBody */
	return msgHeader;
}

void writeMsgBody(FILE *fp, unsigned char *s)
/* Write the string s onto the file fp and transform all control characters
   into C-style. */
{	--s;
	while(*++s)
		if(*s == '\t') fputs("\\t", fp);
		else if(*s == '\n') fputs("\\n\\\n", fp);
		else if(*s == '"') fputs("\\\"", fp);
		else if(*s < ' ') fprintf(fp, "\\%o", *s);
		else fputc(*s, fp);
}

char *incFileNr(char *fileNr)
{	char *p;

	p = fileNr + sizeof(zeroDefFileNr) - 2;
incZyk:
	if(*p == 'Z') {
		*p = '0';
		if(--p >= fileNr)
			goto incZyk;
		else
			fatal(E_tooManyDefFiles);
	}

   	if(*p == '9')
   		*p = 'A';
   	else ++*p;

	return fileNr;
}

FILE *openDefFile(char *head, int errNr)
/* open next message definition file with nme head and error level errNr */
{	FILE *fpX;

	sprintf(msgDefFileStr, msgDefFile, "", msgDefFileNr, ".c");
	incFileNr(msgDefFileNr);
	fpX = Efopen(msgDefFileStr, "wt", NULL);

	if(*head == 'E') 
		message(fpX, msgErrOpening, errNr, head);
	else
		message(fpX, msgOpening, head);

	return fpX;
}

void closeDefFile(FILE *fp)
/* close a message definition file */
{	fputs(msgClosing, fp);
	Efclose(fp, msgDefFileStr, NULL);
}


int internalName(char **str)
{	int i, l;
	char *s;

	i = IN_Nr;
	s = *str;
	while(i--) 
		if(memcmp(s, internal[i], l = strlen(internal[i])) == 0)
			goto found;	/* found */

	return -1;	/* error */

found:
	s += l;
	while(isspace(*s)) ++s;
	*str = s;
	return i;
}

unsigned copyNumber(char *dest, char **src, char nam)
{	unsigned len;
	char *s;

	len = 0;
	s = *src;
	if(isdigit(*s)) {
		while(*s == '0') ++s;
		while(isdigit(*s)) {
			++len;
			*dest++ = *s++;
		}
		if(!len) {
			*dest++ = '0';
			++len;
		}
	}
	*dest++ = nam;
	*dest++ = NUL;
	while(isspace(*s)) ++s;
	*src = s;
	return 1 + len;
}

unsigned translateANSI(char *dest, char *src)
/* translate a ANSI sequence into its escape code representation */
{	int i, j, flag;
	char *s, ansi[32], nam;
	unsigned len;

	len = 0;
	if(s = src = strtok(src, ",")) do {
		while(isspace(*src)) ++src;
		if((i = internalName(&s)) < 0)
			error(E_ANSIToken, src, lineNr);

#define sAnsi(a) stpcpy(ansi, a); break
#define tAnsi(a) nam = a; goto singleArg
#define cAnsi(a) nam = a; goto colourArg
#define aAnsi(a) nam = a; goto argArg
		switch(i) {
			case IN_cls: 		sAnsi("2J");
			case IN_clreol: 	sAnsi("K");
			case IN_home: 		sAnsi("H");
			case IN_save:		sAnsi("s");
			case IN_restore:	sAnsi("u");
			case IN_up:		tAnsi('A');
			case IN_down:		tAnsi('B');
			case IN_left:		tAnsi('D');
			case IN_right:		tAnsi('C');
			singleArg:
				if((i = copyNumber(ansi, &s, nam)) == 1 || i > 15)
					error(E_ANSInumericArg, src, lineNr);
				break;
			case IN_goto:
				if((i = copyNumber(ansi, &s, ';')) == 1 || i > 31
				  || (j = copyNumber(ansi + i, &s, 'H')) == i + 1 || j > 31)
					error(E_ANSIdoubleNumericArg, src, lineNr);
				break;
			default:
				flag = j = 0;	/* first colour is foreground */
				do switch(i) {
					case IN_bright:	aAnsi(1);
					case IN_blink:		aAnsi(5);
					case IN_inverse:	aAnsi(7);
					case IN_underline:	aAnsi(4);
					case IN_hidden:	aAnsi(8);
					case IN_normal:	aAnsi(0);
					case IN_black:		cAnsi(0);
					case IN_red:		cAnsi(1);
					case IN_green:		cAnsi(2);
					case IN_yellow:	cAnsi(3);
					case IN_blue:		cAnsi(4);
					case IN_magenta:	cAnsi(5);
					case IN_cyan:		cAnsi(6);
					case IN_white:		cAnsi(7);
					colourArg:
						nam |= flag? 0x40: 0x30;
						flag = 1;
					argArg:
						if(j < 29) {
							if(nam & 0xf0)
								ansi[j++] = 0x30 | (nam >> 4);
							ansi[j++] = 0x30 | (nam & 0xf);
							ansi[j++] = ';';
							break;
						}
					default:
						error(E_ANSIToken, src, lineNr);
				} while((i = internalName(&s)) >= 0);
				ansi[j - 1] = 'm';
				ansi[j] = NUL;
				break;
		}
		if(*s)
			error(E_ANSIToken, src, lineNr);
		len += 2 + strlen(ansi);
		dest = stpcpy(stpcpy(dest, "\33["), ansi);
	} while(src = s = strtok(NULL, ","));
	return len;
}


void copyDecF(FILE *fout, FILE *fin, int tail)
/* Copy from fin to fout until a /^DCL_MSG(/ is encountered. */
{	static const char trigger[] = TRIGGER;
	char buf[BUF];
	int c;

	do {
		if(!fgets(buf, sizeof(buf), fin))
			return; 	/* EOF */
		c = memcmp(buf, trigger, sizeof(trigger) - 1); /* trigger found */
		/* dump/skip this line */
#ifdef _MICROC_
		/* Micro-C strips the newline character, so a line longer than
		   BUF characters cannot be detected. */
		if(c) {
			fputs(buf, fout);
			fputc('\n', fout);
		}
#else
		do if(c) fputs(buf, fout);
		while(!strchr(buf, '\n') && fgets(buf, sizeof(buf), fin));
#endif
	} while(c || tail);
}

#define tstI(a) if(strcmp(head, IN_PREFIX #a) == 0) internal[IN_##a] = strdup(msgBody); else

main(int argc, char **argv)
{	char *head, *tail, *p, *mapFile;
    FILE *fp, *msgFp, *mapFp, *bakFp;
    unsigned i;
    char fnr[5];

    msgInit();

	if(memcmp(argv[1], "/!", 3) == 0) {	/* use msg!#### instead of msg_#### */
		++argv;
		--argc;	/* /! */
		msgDefFile[padDefFile] = '!';
	}

    if(argc < 3 || argc > 6) hlpScreen();


	/* read internal ANSI strings */
	informative(M_readInternal);
	glbMsgFile = mkFname(argv[1]);
	fp = Efopen(glbMsgFile, "r", extMsg);	/* global message file */
	lineNr = 0;
	readMsgBody(fp);	/* skip header */

    while(head = readMsgHeader(fp))
#ifndef _MICROC_	/* as long Micro-C cannot concat two string literals on the fly */
			tstI(cls)
			tstI(clreol)
			tstI(home)
			tstI(up)
			tstI(down)
			tstI(left)
			tstI(right)
			tstI(goto)
			tstI(save)
			tstI(restore)
			tstI(bright)
			tstI(blink)
			tstI(inverse)
			tstI(underline)
			tstI(hidden)
			tstI(normal)
			tstI(black)
			tstI(red)
			tstI(green)
			tstI(yellow)
			tstI(blue)
			tstI(magenta)
			tstI(cyan)
			tstI(white)
#else
if(strcmp(head, "I_cls") == 0) internal[IN_cls] = strdup(msgBody); else
if(strcmp(head, "I_clreol") == 0) internal[IN_clreol] = strdup(msgBody); else
if(strcmp(head, "I_home") == 0) internal[IN_home] = strdup(msgBody); else
if(strcmp(head, "I_up") == 0) internal[IN_up] = strdup(msgBody); else
if(strcmp(head, "I_down") == 0) internal[IN_down] = strdup(msgBody); else
if(strcmp(head, "I_left") == 0) internal[IN_left] = strdup(msgBody); else
if(strcmp(head, "I_right") == 0) internal[IN_right] = strdup(msgBody); else
if(strcmp(head, "I_goto") == 0) internal[IN_goto] = strdup(msgBody); else
if(strcmp(head, "I_save") == 0) internal[IN_save] = strdup(msgBody); else
if(strcmp(head, "I_restore") == 0) internal[IN_restore] = strdup(msgBody); else
if(strcmp(head, "I_bright") == 0) internal[IN_bright] = strdup(msgBody); else
if(strcmp(head, "I_blink") == 0) internal[IN_blink] = strdup(msgBody); else
if(strcmp(head, "I_inverse") == 0) internal[IN_inverse] = strdup(msgBody); else
if(strcmp(head, "I_underline") == 0) internal[IN_underline] = strdup(msgBody); else
if(strcmp(head, "I_hidden") == 0) internal[IN_hidden] = strdup(msgBody); else
if(strcmp(head, "I_normal") == 0) internal[IN_normal] = strdup(msgBody); else
if(strcmp(head, "I_black") == 0) internal[IN_black] = strdup(msgBody); else
if(strcmp(head, "I_red") == 0) internal[IN_red] = strdup(msgBody); else
if(strcmp(head, "I_green") == 0) internal[IN_green] = strdup(msgBody); else
if(strcmp(head, "I_yellow") == 0) internal[IN_yellow] = strdup(msgBody); else
if(strcmp(head, "I_blue") == 0) internal[IN_blue] = strdup(msgBody); else
if(strcmp(head, "I_magenta") == 0) internal[IN_magenta] = strdup(msgBody); else
if(strcmp(head, "I_cyan") == 0) internal[IN_cyan] = strdup(msgBody); else
if(strcmp(head, "I_white") == 0) internal[IN_white] = strdup(msgBody); else
#endif
			;

	Efclose(fp, glbMsgFile, extMsg);


	informative(M_checkInternal);
	i = IN_Nr;
	while(i--)
		if(internal[i] == NULL)
			error(E_internalNameMissing);

	/* read message file & create the message definition files */
	informative(M_createDefFiles);
	msgFile = mkFname(argv[2]);
	fp = Efopen(msgFile, "r", extMsg);
	if(argc > 5) {
		informative(M_createDecFile);
		unlink(strcat(mapFile = mkFname(argv[5]), extBak));
		if(rename(argv[5], mapFile))
			error(E_renDecFile, argv[5]);
	/* copy leading part of the msg declaration file */
		copyDecF(	mapFp = Efopen(argv[5], "w", NULL),
					bakFp = Efopen(mapFile, "r", NULL), 0);
	}
	else {
		warning(W_decFile);
		mapFp = NULL;
	}
	lineNr = 0;
	readMsgBody(fp);	/* skip header */

	while(head = readMsgHeader(fp)) {
		if(mapFp) fprintf(mapFp, "%s%s);\n", TRIGGER, head);
		msgFp = openDefFile(head, errNr);
		p = (tail = msgBody) - 1;
		while(p = strchr(p + 1, '%')) {	/* specil format spec */
			if(p[1] == '%') {	/* skip %% */
				++p;
				continue;
			}
			if(p[1] != '-' || p[2] != '%') {	/* mayhap %#-% */
				i = 0;
				do {
					while(isdigit(*++p))
						i = i * 10 + *p - '0';
				} while(*p == '-');
				if(*p == '%' && p[-1] == '-') /* special %#-% */
					p += i;
				continue;
			}

			/* %-% found => emit previous text & construct the ANSI esc codes */
			if(p != tail) {	/* emit previous text */
				*p = NUL;
				writeMsgBody(msgFp, tail);
				tail = p;
			}
			if((p = strchr(p + 3, '%')) == NULL)
				error(E_encloserOpen, lineNr);
			*p = NUL;
			i = translateANSI(tail, tail + 3);
			fprintf(msgFp, "%%%u-%%", i);
			writeMsgBody(msgFp, tail);
			tail = p + 1;
		}
		if(*tail)
			writeMsgBody(msgFp, tail);
		closeDefFile(msgFp);
	}
	Efclose(fp, msgFile, extMsg);
/* copy trailing part of the msg declaration file */
	if(mapFp) {
		copyDecF(mapFp, bakFp, 1);
		Efclose(bakFp, mapFile, NULL);
		Efclose(mapFp, argv[5], NULL);
	}

	/* create the compiler response file */
	if(argc > 3) {	/* response file */
		informative(M_createCompRsp);
		fp = Efopen(argv[3], "wt", NULL);
		for(memcpy(fnr, zeroDefFileNr, sizeof(zeroDefFileNr));
			memcmp(fnr, msgDefFileNr, sizeof(zeroDefFileNr));
		    incFileNr(fnr))
				fprintf(fp, msgDefFile, "", fnr, ".c\n");
		Efclose(fp, argv[3], NULL);
	}
	else warning(W_compRsp);

	/* create the librarian response file */
	if(argc > 4) {	/* response file */
		informative(M_createLibRsp);
		fp = Efopen(argv[4], "wt", NULL);
		head = "+ ";
		for(memcpy(fnr, zeroDefFileNr, sizeof(zeroDefFileNr));
			memcmp(fnr, msgDefFileNr, sizeof(zeroDefFileNr));
		    incFileNr(fnr))
				fprintf(fp, msgDefFile, head, fnr, ""), head = ".obj &\n+ ";
		if(*head == '.') fputs(".obj\n", fp);
		Efclose(fp, argv[4], NULL);
	}
	else warning(W_libRsp);

#ifndef _MICROC_
	return 0;	/* standard for Micro-C */
#endif
}
