/*
    Source of the message handling library
    Copyright (C) 1995  Steffen Kaiser

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   
*/
/* $RCSfile: msgprntf.c $
   $Locker: ska $	$Name: version2 $	$State: Exp $

    Supported compilers:
    	+ updated version of Micro-C 3.13
    	+ Borland C++ v3.1 and v4.0

	Emit message strings:
	 special syntaxes within format string:
		+ %#-%XXX	(# == decimal; is number of: XXX == any character)
			=> emit ANSI Escape sequence; ESC is written automatically
		+ %fmt		(fmt == normal format specification)
			=> is replaced by the next argument, which has the specified 
			   type
		+ %#%fmt	(# == decimal; fmt == format specification as printf() would accept)
			=> is replaced by the #th argument, which has the specified
			   type; first argument with # == 1
		+ %-#%frm	(# == decimal; is number of the skipped argument)
			=> defines, that the #th argument has this type, but is ignored
			   fully; first argument with # == 1

		+ there can be multiple statements with the same number for:
		  %#%fmt and %-#%fmt, but always the last is used for defining
		  the argument size.
		+ the "next argument" is the first undefined argument, e.g.:
		  "...%1%s...%s...%2%s"
		  ==> both %s and %2%s expand to the string given as the 2nd argument
		  whereas:
		  "...%s..%2%s..%d"
		  ==> %s is 1st, %2%s is 2nd, %d is 3rd
		  and
		  "...%2%s...%s..%d"
		  ==> %2%s is 2nd, %s is 1st, %d is 3rd
		  and
		  "..%1%s..%1%p..%1%u..%1%c..%s"
		  ==> in near data model: %1%s, %1%p, %1%u,& %1%c will define the
		      first argument as 2 bytes wide
		  ==> in far data model: 
		  		arg2-byte 3
		  		arg2-byte 2
		  		arg2-byte 1 \
		  		arg2-byte 0  \(3)
		  		arg1-byte 3  /    \
		  		arg1-byte 2	/      \(2)
		  		arg1-byte 1 \ (1)  /
		  		arg1-byte 0 /     /

		  		(1) used by %1%u,& %1%c
		  		(2) used by %1%s,& %1%p
		  		(3) used by %s, because the 2 bytes-wide definition for
		  		    argument 1 is define last (with %1%c).


	 examples: (double quotes shall mark the C string)
	 	"disk %c: is full"

	 	Three-argument message:
	 	"Disk %c: has the label '%s' and is %s"
		If this message should be altered to:
		"Disk %c: is %s"	the second argument is given to the message() function,
							but should be avoided.
		=> correct: "Disk %c: is %3%s%-2%s"
		                             ^^^^^ defines argument 2 as char*, but is replaced by ""
		                         ^^^^ is replaced by the third argument

		"Disk with label '%4-%\33[7m%s%4-%\33[0m' is full"
					      |   |     | |   ^^^^^^ ANSI escape sequence
						  |   |     | ^^^^ define, that next 4 characters are ANSI
						  |   |     ^^ normal format string to emit a char*
					      |   ^^^^^^ ANSI escape sequence
						  ^^^^ define, that next 4 characters are ANSI



	Algorithm:
	The normal printf() function determines the length of each argument
	by the supplied format string. When the arguments may differ in
	length, it is impossible to calculate the address of an unread argument.
	There are three ways to come along with that restriction:
	1) Write a brand-new printf() function.
	   That would require to much overhead right now and is non-portable,
	   because of the usage of floating point numbers.
	2) Define, that all arguments must have the same length.
	   Hmm, that's sounds unacceptable, because the only reason to
	   use a non-Micro-C compiler is to use its advantages over
	   Micro-C, e.g. floating point arethmetics, far data model.
	3) Prescan the format string and do the swapping with that knowledge.
	   Seems the best here, therefore it is implemented. (Note: The limit
	   of one argument per message was refused, which would be another way.)

	How to realize the argument swapping?
	3.1) Swap the space allocated for ach argument within the stack.
	     There are optimizations, which can re-use that arguments on the
	     stacks -> refused.
	3.2) Duplicate the necessary space, then 3.1).
	     To use dynamic memory is not wise here, therefore the allocated
	     memory had to be within the stack or, by the defining the maximal
	     amount of arguments, is the worst-case (MAXARG * sizeof(long double))
	     as a static buffer. Refused because:
	     a) Maybe, there is a longer type than long double somewhere.
	     b) The format string must be readible by the original
	        printf() function. The argument swapping itself could be
	        stored away from the format string, but not the colour
	        support.
	     Therefore, the format string itself had to be copied
	     => not acceptable without dynamic memory.
	3.3) Call printf() for each argument.
	     Two more advantages:
	     a) Allow to add new format specifications, e.g. %b for consistency
	        between Micro-C and non-Micro-C.
	     b) Handle more than 132 bytes long output in Micro-C.



	Special notes to the Micro-C implementation:
		In Micro-C all arguments have the same length (2 bytes). Therefore
		it is not necessary to calculate the addresses through the format
		string.



	Known bugs:
		+ on non-Micro-C: the %b format was added for consistency purpose
		  to Micro-C. It supports (unsigned) and (unsigned long). The
		  size modifier 'l' must immediatly preceed the 'b'. The "blank
		  when zero" option is not implemented. There is no alternate form.
		+ on Micro-C: the %s format was re-implemented. Therefore, to enable
		  the "Fill leading space with zeros" option the zro must immediatly
		  follow the '%', e.g. "%0123s".

   $Log: msgprntf.c $
   Revision 1.1  1995/05/26 22:53:58  ska
   Initial revision

   Revision 1.7  1995/05/26 20:25:12  ska
   bugfix: argument swapping in non-Micro-C part.

   Revision 1.6  1995/05/16 17:58:06  ska
   bugfix: sizeof calculation for floats.
   bugfix: filler in length calculation of %s for Micro-C.

   Revision 1.5  1995/04/21 04:35:46  ska
   version 1.0

   Revision 1.4  1995/04/21 04:11:29  ska
   chg: directory structure.

   Revision 1.3  1995/04/19 09:06:46  ska
   chg: no format string error reporting

   Revision 1.2  1995/04/11 17:57:54  ska
   Micro-C hack-in.

   Revision 1.1  1995/04/10 23:20:00  ska
   Initial revision

*/

#include <stdio.h>
#ifndef _MICROC_
#include <string.h>
#include <stdarg.h>		/* definition of type va_list */
#else	/* Micro-C */
#include <file.h>
#endif
#include <freedos.h>
#include <msglib.h>

/* !!! ska better no error report 
   ANSI defines that an corrupt format string is *implementor defined*;
   so I define: ignore it most times, but return with EOF if it's a 
   fatal error.

   int __formatStringError(const char *str); */

#ifdef _MICROC_
int fputmc(int c, int rep, FILE *fp);
#endif

#define MAXARG 32		/* maximal number of arguments */
#define FL_DASH ((unsigned)1 << (sizeof(unsigned) * 8 - 1))
#define FL_MISC (FL_DASH >> 1)

#ifdef USE_ISADEVICE
#define isatty isadev
#endif

#ifndef lint
static char const rcsid[] = 
	"$Id: msgprntf.c 1.1 1995/05/26 22:53:58 ska Exp ska $";
#endif

static int ansiTest = ANSI_TEST;

int getAnsiTest(void)
{	return ansiTest;	}
void setAnsiTest(int flag)
{	ansiTest = flag;	}


	/* 
	   arg[] holds the offset of the arguments as they are supplied on the
	         caller's function parameter list.

	   argMap[] specifies, that the idx'th argument of the format string
	   		is the argMap[idx]'th argument within the arg[] array.
	   		This is necessary for the "next arguments" without the
	   		preceeding %# and for unused arguments.

	*/


	/* The colour support

		bases upon ANSI to allow to use the stream functions. Further
		implementations may also differ between ANSI/direct/no colour.

		ANSI is necessary for redirecting (remote access), direct for 
		colour without an ANSI driver, and none for redirecting into
		a file.

		To differ between ANSI/non-ANSI the current implementation uses
		the global variable ansiTest:
			ANSI_TEST: The function will use ANSI only , if the output stream
				is connected to a CON device.
			ANSI_ALLOW: Use ANSI ever.
			ANSI_DENY: Use ANSI never.


		The test is made with the isatty() function.
	
	*/

	/* Special for Micro-C

	   + argMap is used to indicate used arguments to skip for the
	     "next argument" calculation.

	   + The whole part to determine the offsets and the swapping is cut.
	     Therefore, the format error detection must be moved into the
	     second while loop. 

	   + The s-format is not handled by printf().

	   + printf() is not called by vprintf(), but by duplicating the
	     argument.

	*/


#ifdef _MICROC_
#define skfwrite(buf,len,fp) fput(buf,len,fp)	/* item size is always 1 */
#define ARGS_TYPE
#else
#define skfwrite(buf,len,fp) fwrite(buf,1,len,fp)
#define ARGS_TYPE char
#endif

int skprintf(FILE *fout, const char *fmt, const unsigned ARGS_TYPE *args)
/* fout: output stream
   fmt: enhanced format string
   args: pointer to the first argument

   return: number of written characters or EOF on failure
*/
{	char *s, *p;	/* pointer within the format string */
	int idx, anr, nr;	/* argument indexes */
	unsigned flags, size, length;
	int ansi;		/* 0: skip ANSI sequences; !0: put them into the stream */
	unsigned argMap[MAXARG];	/* arg idx ==> swapped to arg argMap[idx] */
#ifndef _MICROC_
	unsigned arg[MAXARG], *as;	/* points to next argument slot */

	memset(arg, 0, sizeof(arg));
#endif
	if((ansi = ansiTest) == ANSI_TEST)
		ansi = isatty(fileno(fout))? ANSI_ALLOW: ANSI_DENY;
	ansi -= ANSI_DENY;

#ifndef _MICROC_
	/* 1. Create the argument lengths */
	p = (char*)fmt - 1;
	as = argMap;
	idx = 0;

	while(p = strchr(s = p + 1, '%')) {	/* at least one special */
		anr = idx;
		flags = size = nr = 0;
		do switch(*++p) {
			case '%':
				if(flags == FL_DASH && p[-1] == '-') { /* special %#-%XXXX ANSI */

					/* there is no test, if the string is less than the 
					   requested length. This will allow '\0's (for what
					   reason ever). */
					p += nr;
					goto noArg;	/* did not consumed an arg */
				}
				else if(p[-1] == '%')	/* special %% */
					goto noArg;
				else if((flags & ~FL_DASH) == 0) {	/* special %#% or %-#% => argument switching or suppressing */
					anr = nr - 1;
					continue;
				}

				/* fall through to quirk */
			default:
				/* return __formatStringError(p); */	/* format string quirk */
				/* !!! ska better no error reporting */
				goto noArg;

			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				nr = nr * 10 + *p - '0';
				continue;

			case '-': flags |= FL_DASH; continue;

			case ' ': case '+': case '#': case '.':
				flags |= FL_MISC;	/* these are valid flags, but for no use now */
				continue;

			case '*': ++flags; continue;	/* This flag indicates an additional (int) item used up by this argument */

			case 'N': case 'F': case 'l': case 'h': case 'L':	/* size modificating prefixes */
				size = *p;
				continue;

			case 'd': case 'i': case 'o': case 'u':
			case 'x': case 'X':
/* Micro-C does not accept #if 0 */
#ifdef NICHT_UEBERGEHEN		/* automatically promoted to  int */
				if(sizeof(short) != sizeof(int)) {	/* I hope, your compiler removes unreachable code */
					if(size == 'h')	{
						size = sizeof(short);
						goto Arg;
					}
				}
#endif
				if(sizeof(long) != sizeof(int)) {	/* I hope, your compiler removes unreachable code */
					if(size == 'l')	{
						size = sizeof(long);
						goto Arg;
					}
				}
				/* default: sizeof(int) */

			case 'c': size = sizeof(int); goto Arg;

			case 'f': case 'e': case 'E': case 'g': case 'G':
/* Micro-C does not accept #if 0 */
#ifdef NICHT_UEBERGEHEN			/* float is automatically promoted to double */
				size = size == 'l'? sizeof(double):
					   size == 'L'? sizeof(long double): sizeof(float);
#else
				size = size == 'L'? sizeof(long double): sizeof(double);
#endif
				goto Arg;
			case 's': case 'p': case 'n':
				if(sizeof(void*) == sizeof(void near*)) {	/* I hope, your compiler removes unreachable code */
					size = size == 'F'? sizeof(void far*): sizeof(void*);
				} else {
					size = size == 'N'? sizeof(void near*): sizeof(void*);
				}
				goto Arg;

		} while(1);

Arg:	/* an argument was used */

		if(anr > MAXARG-1) 	/* argument anr is out of bound */
			/* return __formatStringError(p); */	/* argument amount overflow */
			return EOF;	/* try to avoid rubish */

		/* the size of the argument and its additional (int) arguments */
		arg[*as++ = anr] = size + (flags & ~(FL_DASH | FL_MISC)) * sizeof(int);
		/*  ^^^^^^^^^^^       current argument is mapped to anr */

		if(arg[idx] != 0)	/* advance to next argument */
			while(++idx < MAXARG && arg[idx]);
			/* allow idx == MAXARG, for axactly MAXARG arguments */

noArg:
	}

	/* 2. Evaluate argument offsets; first argument starts with 0 */
	/* evaluate offsets of the arguments in the original order */
	for(size = idx = 0, anr = -1; ++anr < MAXARG;) {
		if(nr = arg[anr]) {
			if(idx) /* with a missing argument, offset not calculatable */
				/* return __formatStringError(fmt); */	/* a missing argument */
				return EOF;	/* unable to evaluate the argument address */
		}
		else idx = 1;

		arg[anr] = size;
		size += nr;
	}

	/* 2.1. map indexes to offsets, so argMap[] points correctly */
	while(--as >= argMap) {
		*as = arg[*as];
	}
#else	/* Micro-C */

	memset(argMap, 0, sizeof(argMap));	/* no argument already used */
#endif


	/* 3. Emit message by using the "original" printf */
	/* here should be no errors anymore */

	p = (char*)fmt - 1;
	length = idx = 0;

	while(p = strchr(s = p + 1, '%')) {	/* at least one special */

		if(s != p) {	/* emit constant text */
			/* the string length is known, so fputs() need
			   not calculateit again; need not to place an
			   '\0' at the end */
			if(skfwrite(s, nr = p - s, fout) != nr)
				return EOF;	/* some error occured */
			length += nr;
			s = p;
		}

		anr = idx;
		flags = size = nr = 0;
		do switch(*++p) {
			case '%':
				if(p[-1] == '-') { /* special %#-%XXXX ANSI */
					if(ansi) {
						/* again: skfwrite() need not recalculate string length
						          and no trailing '\0' */
						if(skfwrite(p + 1, nr, fout) != nr)
							return EOF;	/* some error occured */
						length += nr;
					}
					p += nr;
					goto noArg1;	/* did not consumed an arg */
				}
				else if(p[-1] == '%') {	/* special %% */
					fputc('%', fout);
					++length;
					goto noArg1;
				}
				else if((flags & ~FL_DASH) == 0) {	/* special %[-]#% => argument switching/suppressing */
#ifndef _MICROC_
					anr = nr - 1;
					/* nr = 0;	not necessary, because field width is generated by printf() */
#else
					if((anr = nr - 1) > MAXARG - 1)
						/* return __formatStringError(p); */
						return EOF;
					nr = 0;	/* for %s */
#endif
					s = p;	/* save start of format spec */
					continue;
				}

			default:
#ifdef _MICROC_
				/* return __formatStringError(s); */
				break;
#else
				break;	/* it's for robustness */
#endif

			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				nr = nr * 10 + *p - '0';	/* for ANSI enclosures only */
				continue;

			case '-': flags |= FL_DASH; continue;

			/* The misc flags are of no interest anymore */

#ifndef _MICROC_
			case 'd': case 'i': case 'o': case 'u':
			case 'x': case 'X':
			case 'c': 
			case 'f': case 'e': case 'E': case 'g': case 'G':
			case 's': case 'p': case 'n':
#else
			case 'd': case 'o': case 'u':
			case 'x': 
			case 'c': 
			case 'b':
#endif

				if(flags & FL_DASH)	/* do not display this argument */
					goto Arg1;

				size = p[1];	/* create the sub-string holding the format */
				p[1] = '\0';	/* spec only */
#ifndef _MICROC_
				if((nr = vfprintf(fout, s, (va_list)(args + argMap[anr]))) == EOF)
#else
				if((nr = fprintf(fout, s, args[-anr])) == EOF)
#endif
					return EOF;
				length += nr;
				p[1] = size;

				goto Arg1;

#ifdef _MICROC_
			case 's':	/* special string handling */
				if(flags & FL_DASH)	/* do not display this argument */
					goto noArg1;

				size = strlen(args[-anr]);
				if(nr)
					if(nr > size) {	/* fitting */
						length += size - nr;	/* filler */
						if(flags & FL_DASH) { /* left justify */
							if(skfwrite(args[-anr], size, fout) != size
							  || fputmc(' ', size - nr, fout))
								return EOF;	/* some error occured */
							goto emitted;
						}
						if(fputmc(s[1] == '0'? '0': ' ', size - nr, fout))
							return EOF;
					}
					else size = nr;	/* shrink */

				if(skfwrite(args[-anr], size, fout) != size)
					return EOF;	/* some error occured */
			emitted:
				length += size;
				goto Arg1;
#else
			case 'b':	/* for consistency to Micro-C */
				if(flags & FL_DASH)	/* do not display this argument */
					goto noArg1;

				{	char buf[sizeof(unsigned long) * 8 + 1];	/* assume long */
					char *bufp;
					unsigned long it;

					bufp = buf;
					it = *(unsigned long*)(args + argMap[idx]);
					if(p[-1] != 'l') it &= 0xffff;

					if(it) 
						do *bufp++ = it & 1? '1': '0';
						while(it >>= 1);
					else *bufp++ = '0';

					*bufp = '\0';
					*p = 's';
					size = p[1];
					p[1] = '\0';
					fprintf(fout, s, strrev(buf));
					p[1] = size;
					*p = 'b';
				}
				goto Arg1;
#endif


		} while(1);

Arg1:
#ifndef _MICROC_
		++idx;
#else	/* "next argument" calculation */
		argMap[anr] = 1;
		while(argMap[idx])
			if(++idx >= MAXARG)
				/* __formatStringError(p); */
				return EOF;
#endif
noArg1:

	}

	/* 4. Emit any trailing string */
	if(*s)	{ /* There're some chars left */
		if(skfwrite(s, nr = strlen(s), fout) != nr)
			return EOF;	/* some error occured */
		length += nr;
	}

	return length;
}

#ifdef _MICROC_

register int fd_fprintf(FILE *fp, const char *msg)	/* need two args */
{
	asm {
		shl ax, 1	; amount of bytes used by all arguments
		lea bx, 2[bp]	; address below last arg
		add bx, ax	; address of file pointer
		push Word Ptr [bx]	; 1st arg: file pointer
		dec bx
		dec bx		; address of format string
		push Word Ptr [bx]	; 2nd arg: formt string
		dec bx
		dec bx		; address of 1st printf()''s arg
		push bx		; 3rd arg: addrss of 1st arg for printf()
		call _skprintf
		add sp, 6
	}
}

#else

int fd_fprintf(FILE *fp, const char *msg, ...)
{	return skprintf(fp, msg, (const unsigned ARGS_TYPE*)(msg + 1));	}

#endif

